/*
 * com.micheldalal.x10 library.  Set of X10 Home Control classes.
 *
 * Copyright (C) 2001-2003  Michel Dalal.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You may also find a copy of the licence at
 * <a href="http://www.gnu.org/copyleft/lesser.html">http://www.gnu.org/copyleft/lesser.html</a>
 */

package com.micheldalal.x10.test;

import java.util.Enumeration;
import javax.comm.CommPortIdentifier;
import javax.comm.NoSuchPortException;
import javax.comm.PortInUseException;
import javax.comm.SerialPort;
import com.nevelex.app.args.ArgumentConsumer;
import com.nevelex.app.args.ArgumentCommand;
import com.nevelex.app.args.ArgumentException;
import com.nevelex.app.args.ArgumentProcessor;
import com.nevelex.app.args.InvalidArgumentException;
import com.nevelex.app.args.MissingArgumentException;
import com.micheldalal.x10.CM17A;

/**
 * Copyright (C) 2001-2003 Michel Dalal
 */
/**
 * Sample class to show how to use X10 communication stuff.
 * <p>
 * On Linux, tested with <a href="http://www.rxtx.org/">rxtx</a> version 2.0-5.
 * <p>
 * This particular test program uses the <a href="http://nevelex.com/products/NevArgs/">Nevelex Arguments API</a> for
 * command-line argument processing.
 *
 * @version 1.0 07/05/2001: Initial Version.
 * @version <br>1.0.1 10/20/2003: Comment updates.
 * @author Michel Dalal
 */
public class X10Test implements ArgumentConsumer
{
	/** The default serial port. */
	public final static String LINUX_DEFAULT_PORT = "/dev/ttyS0";
	public final static String WINDOWS_DEFAULT_PORT = "COM1";
	
	/** The set of available run modes. */
	private final static int RUN_MODE_UNSET=0;
	private final static int RUN_MODE_OFF=1;
	private final static int RUN_MODE_ON=2;
	private final static int RUN_MODE_DIM=3;
	private final static int RUN_MODE_BRIGHTEN=4;
	
	/** The desired port name. */
	private String portName;
	
	/** The reference to the serial port. */
	private SerialPort serialPort;

	/** The run mode. */
	private int runMode;
	
	/** The house code. */
	private char houseCode;
	
	/** The unit id. */
	private int unitID;
	
	/** The percentage change value. */
	private int percent;
	
	/**
	 * Constructor for this object just initializes into a house code of 'A', a unit id of 1
	 * and the run mode is unset.
	 */	
	public X10Test()
	{
		this.portName = null;
		this.serialPort = null;
		this.runMode = RUN_MODE_UNSET;
		this.houseCode='A';
		this.unitID=1;
		this.percent=0;
	}

	/**
	 * Attemps to find the desired port by name.
	 *
	 * @return the portID if found or null if it is not available.
	 * @exception NoSuchPortException if unable to locate the port.
	 */
	private CommPortIdentifier getPortID() throws NoSuchPortException
	{
		return CommPortIdentifier.getPortIdentifier(portName);
	}

	/**
	 * Attempts to turn on or off the transiever module.
	 * D1 ON           01010000   00000000
	 * D1 OFF          01010000   00100000
	 *
	 * @exception PortInUseException throw if unable to open the port.
	 */
	private void doIt() throws PortInUseException
	{
		CommPortIdentifier portID;
		CM17A cm17A;

		try
		{
			portID = getPortID();
			serialPort = (SerialPort)portID.open(portName,2000);
			cm17A = new CM17A(serialPort);
			switch (runMode)
			{
				case RUN_MODE_OFF:
				case RUN_MODE_ON:
					// Set the state of the device.
					cm17A.setState(houseCode,unitID,runMode == RUN_MODE_ON);
					break;
				case RUN_MODE_DIM:
				case RUN_MODE_BRIGHTEN:
					// Adjust the brightness for the lamp modules in a house.
					cm17A.adjustBrightness(houseCode,runMode == RUN_MODE_BRIGHTEN,percent);
					break;
			}
			serialPort.close();
		}
		catch (NoSuchPortException nspe)
		{
			System.err.println("Port not found: " + portName);
		}
	}

	/////////////////////////////////////////////////////////////////////////////////
	// Application related methods and inner classes.
	/////////////////////////////////////////////////////////////////////////////////
	/**
	 * Command pattern class implementing the Port Argument.
	 */
	class PortArgumentCommand implements ArgumentCommand
	{
		/**
		 * Consumes the current command-line argument.
		 * The element in the args array at the offset will be the argument itself.
		 * This method is the equivalent of the execute method in most command interfaces in the command pattern.
		 *
		 * @param args the command line argument array.
		 * @param offset the offset in to the command line argument array where the argument exists.
		 * @return the offset for the next argument in the command.
		 * @exception InvalidArgumentException thrown if the argument does not have all of its required options.
		 */
		public int consumeArgument(java.lang.String[] args,int offset) throws InvalidArgumentException
		{
			offset++;
			if (args.length>offset)
			{
				portName = args[offset++];
			}
			else
			{
				throw new InvalidArgumentException("Port argument requires a <serial_port>.",offset);
			}
			return offset;
		}
	}

	/**
	 * Command pattern class implementing the Run Mode Argument.
	 */
	class RunModeArgumentCommand implements ArgumentCommand
	{
		/**
		 * Consumes the current command-line argument.
		 * The element in the args array at the offset will be the argument itself.
		 * This method is the equivalent of the execute method in most command interfaces in the command pattern.
		 *
		 * @param args the command line argument array.
		 * @param offset the offset in to the command line argument array where the argument exists.
		 * @return the offset for the next argument in the command.
		 * @exception InvalidArgumentException thrown if the argument does not have all of its required options.
		 */
		public int consumeArgument(java.lang.String[] args,int offset) throws InvalidArgumentException
		{
			if (args[offset].equals("-off"))
			{
				offset++;
				runMode = RUN_MODE_OFF;
			}
			else if (args[offset].equals("-on"))
			{
				offset++;
				runMode = RUN_MODE_ON;
			}
			else
			{
				if (args[offset++].equals("-b"))
				{
					runMode = RUN_MODE_BRIGHTEN;
				}
				else
				{
					runMode = RUN_MODE_DIM;
				}
				
				if (offset<args.length)
				{
					try
					{
						percent = Integer.parseInt(args[offset++]);
						if (percent<5||percent>100)
						{
							throw new NumberFormatException();
						}
					}
					catch (NumberFormatException e)
					{
						throw new InvalidArgumentException("Dim/Brighten argument requires a numeric <percent> between 5 and 100 inclusive.",offset);
					}
				}
				else
				{
					throw new InvalidArgumentException("Dim/Brighten argument requires a <percent> between 5 and 100 inclusive.",offset);
				}
			}
			return offset;
		}
	}

	/**
	 * Used to specify the house code.
	 */
	class HouseCodeArgumentCommand implements ArgumentCommand
	{
		/**
		 * Consumes the current command-line argument.
		 * The element in the args array at the offset will be the argument itself.
		 * This method is the equivalent of the execute method in most command interfaces in the command pattern.
		 *
		 * @param args the command line argument array.
		 * @param offset the offset in to the command line argument array where the argument exists.
		 * @return the offset for the next argument in the command.
		 * @exception InvalidArgumentException thrown if the argument does not have all of its required options.
		 */
		public int consumeArgument(java.lang.String[] args,int offset) throws InvalidArgumentException
		{
			offset++;
			if (offset<args.length)
			{
				houseCode = Character.toUpperCase(args[offset++].charAt(0));
				if (houseCode<'A'||houseCode>'P')
				{
					throw new InvalidArgumentException("House code argument requires a <house_code> between 'A' and 'P' inclusive.",offset);
				}
			}
			else
			{
				throw new InvalidArgumentException("House code argument requires a <house_code>. ('A' to 'P')",offset);
			}
			return offset;
		}
	}

	/**
	 * Used to specify the unit id.
	 */
	class UnitIDArgumentCommand implements ArgumentCommand
	{
		/**
		 * Consumes the current command-line argument.
		 * The element in the args array at the offset will be the argument itself.
		 * This method is the equivalent of the execute method in most command interfaces in the command pattern.
		 *
		 * @param args the command line argument array.
		 * @param offset the offset in to the command line argument array where the argument exists.
		 * @return the offset for the next argument in the command.
		 * @exception InvalidArgumentException thrown if the argument does not have all of its required options.
		 */
		public int consumeArgument(java.lang.String[] args,int offset) throws InvalidArgumentException
		{
			offset++;
			if (offset<args.length)
			{
				try
				{
					unitID = Integer.parseInt(args[offset++]);
					if (unitID<1||unitID>16)
					{
						throw new NumberFormatException();
					}
				}
				catch (NumberFormatException e)
				{
					throw new InvalidArgumentException("Unit ID argument requires a numeric <unit_id> between 1 and 16 inclusive.",offset);
				}
			}
			else
			{
				throw new InvalidArgumentException("Unit ID argument requires a <unit_id>. (1 to 16)",offset);
			}
			return offset;
		}
	}

	/**
	 * The method is called by the ArgumentProcessor <b>before</b> any the arguments
	 * have been processed by the application.  It checks the state of the
	 * consumer.  If the consumer is not in a valid state, a IllegalStateException
	 * will be thrown.  This method may be used to register all
	 * {@link ArgumentCommand ArgumentCommand}s
	 * with the {@link ArgumentProcessor ArgumentProcessor}.
	 *
	 * @exception IllegalStateException thrown if the application is not in the desired state.
	 * @see ArgumentCommand
	 * @see ArgumentProcessor
	 */
	public void initArgumentState() throws IllegalStateException
	{
	}

	/**
	 * The method is called by the ArgumentProcessor <b>after</b> all the arguments
	 * have been processed by the application.  It checks the state of the
	 * consumer.  If the consumer is not in a valid state, a MissingArgumentException
	 * will be thrown.
	 *
	 * @exception MissingArgumentException thrown if the argument list did not contain all the required arguments.
	 */
	public void checkArgumentState() throws MissingArgumentException
	{
		if (portName == null)
		{
			if (System.getProperty("os.name").equalsIgnoreCase("linux"))
			{
				portName = LINUX_DEFAULT_PORT;
			}
			else if (System.getProperty("os.name").toLowerCase().startsWith("win"))
			{
				portName = WINDOWS_DEFAULT_PORT;
			}
			else
			{
				throw new MissingArgumentException("Please specify a port name.");
			}
		}
		if (runMode == RUN_MODE_UNSET)
		{
			throw new MissingArgumentException("Please specify a mode.");
		}
	}

	/**
	 * This method handles processing of command-line arguments.
	 *
	 * @param args the command line arguments.
	 * @exception ArgumentException thrown when there are command-line argument problems.
	 */
	private void init(String []args) throws ArgumentException
	{
		ArgumentCommand argumentCommand;

		ArgumentProcessor.setArgumentConsumer(this);
		argumentCommand = new PortArgumentCommand();
		ArgumentProcessor.registerArgumentCommand("-p",argumentCommand);
		argumentCommand = new RunModeArgumentCommand();
		ArgumentProcessor.registerArgumentCommand("-on",argumentCommand);
		ArgumentProcessor.registerArgumentCommand("-off",argumentCommand);
		ArgumentProcessor.registerArgumentCommand("-d",argumentCommand);
		ArgumentProcessor.registerArgumentCommand("-b",argumentCommand);
		argumentCommand = new UnitIDArgumentCommand();
		ArgumentProcessor.registerArgumentCommand("-u",argumentCommand);
		argumentCommand = new HouseCodeArgumentCommand();
		ArgumentProcessor.registerArgumentCommand("-h",argumentCommand);
		ArgumentProcessor.processArguments(args);
	}
	
	/**
	 * Displays the usage help.
	 */
	private static void usage()
	{
		System.err.println("Application to communicate via the FireCracker(TM) Computer Interface (CM17A)");
		System.err.println("to the Wireless Link Module(TM) (a.k.a. Transiever Module).\n");
		
		System.err.println("Usage: X10Test [options...] <mode> [options...]");
		System.err.println("  Modes: (only one mode allowed)");
		System.err.println("\t -off               => Turns off the specified unit id.");
		System.err.println("\t -on                => Turns on the specified unit id.");
		System.err.println("\t -d <percent>       => Dims a light by the specified percentage (should be a");
		System.err.println("\t                       multiple of 5%).");
		System.err.println("\t -b <percent>       => Brightens a light by the specified percentage (should");
		System.err.println("\t                       be a multiple of 5%).");
		System.err.println("  Options:");
		System.err.println("\t -p <serial_port>   => Specifies the serial device.  Defaults to /dev/ttyS0");
		System.err.println("\t                       on Linux.");
		System.err.println("\t -h <house_code>    => The house code to control.  A character value from 'A'");
		System.err.println("\t                       to 'P' inclusive.  Defaults to 'A'.");
		System.err.println("\t -u <unit_id>       => The unit id to control.  A numeric value from 1 to 16");
		System.err.println("\t                       inclusive.  Defaults to 1.");
	}

	/**
	 * Start of the main program.
	 */
	public static void main(String []args)
	{
		X10Test tester;

		tester = new X10Test();
		try
		{
			tester.init(args);
			tester.doIt();
		}
		catch (ArgumentException ae)
		{
			System.err.println("Usage Error: "+ae.getMessage());
			usage();
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}
}

